/*
 * Copyright (C) 1999 Arthur Jerijian
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * In addition, permission is hereby granted, free of charge, to any person
 * obtaining a copy of the software, to deal in the Software according to the
 * terms outlined in the GNU General Public License (the "GPL") in lieu of the
 * licensing terms stated in the above permission notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) OF THE SOFTWARE BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * capture.c: Screen capture management routines
 */

#include <unistd.h>
#include "common.h"
#include "util.h"
#include "stringtable.h"
#include "capture.h"

/* Global variables */

/* Screen capture parameters */
capture_type cap_type = DEFAULT_TYPE;
gint cap_show = DEFAULT_SHOW;
unsigned int cap_delay = DEFAULT_DELAY;

/* ImageMagick image data structures */
Image *magick_image = NULL;
ImageInfo *magick_image_info = NULL;
XImportInfo magick_import_info;

/* Screen capture output window data structures */
GtkWidget *capture_window = NULL;
GtkWidget *capture_pixmap = NULL;

/* Screen capture image data structures for previewing */
GdkImlibImage *capture_image = NULL;
GdkPixmap *capture_gdk_pixmap = NULL;
GdkBitmap *capture_gdk_mask = NULL;

/* File name strings */
GString *capture_temp_filename = NULL;
GString *capture_temp_filename_24 = NULL;
GString *save_filename = NULL;

/* Main capture window callback functions */

gint capture_window_delete_event
(
    GtkWidget *widget,
    GdkEvent *event,
    gpointer data
)
{
    return (TRUE);
}

void capture_window_destroy (GtkWidget *widget, gpointer data)
{
    destroy_capture_image ();
}

/*********************************************************************
 *
 * destroy_magick_image: Destroy the ImageMagick data structures and
 * free the structures' allocated memory.
 *
 * Parameters:
 *   None.
 *
 * Return value:
 *   None.
 */

void destroy_magick_image ()
{
    if (magick_image != NULL)
    {
        DestroyImage (magick_image);
        magick_image = NULL;
    }
    if (magick_image_info != NULL)
    {
        DestroyImageInfo (magick_image_info);
        magick_image_info = NULL;
    }
}

/*********************************************************************
 *
 * destroy_capture_image: Destroy the screen capture preview image
 * structure and free the structure's allocated memory.
 *
 * Parameters:
 *   None.
 *
 * Return value:
 *   None.
 */

void destroy_capture_image ()
{
    if (capture_image != NULL)
    {
        gdk_imlib_destroy_image (capture_image);
        capture_image = NULL;
    }
    if (capture_gdk_pixmap != NULL)
    {
        gdk_pixmap_unref (capture_gdk_pixmap);
    }
    if (capture_gdk_mask != NULL)
    {
        gdk_bitmap_unref (capture_gdk_mask);
    }

}

/*********************************************************************
 *
 * destroy_capture_window: Close the screen capture window and free
 * the window's allocated memory.
 *
 * Parameters:
 *   None.
 *
 * Return value:
 *   None.
 */

void destroy_capture_window ()
{
    if (capture_pixmap != NULL)
    {
        gtk_widget_destroy (capture_pixmap);
        capture_pixmap = NULL;
    }
    if (capture_window != NULL)
    {
        /*
         * Some window managers and desktops (like KDE on Linux
         * and CDE on Solaris 7) trick this program into thinking that
         * the screen capture window doesn't exist when it is actually
         * just minimized. We need to show the window so that it may be
         * properly destroyed.
         */

        gtk_widget_destroy (capture_window);
        capture_window = NULL;
    }
}

/*********************************************************************
 *
 * capture_init: Initialize the screen capture variables and data
 * structures. This function must be called after initializing all
 * the toolkits and before creating the main panel window.
 *
 * Parameters:
 *   None.
 *
 * Return value:
 *   None.
 */

void capture_init ()
{
    pid_t pid;
    GdkVisual *visual;
    GdkColormap *colormap;

    /* Set default screen capture parameters. */
    
    cap_type = DEFAULT_TYPE;
    cap_show = DEFAULT_SHOW;
    cap_delay = DEFAULT_DELAY;

    /* Initialize all data structure pointers to NULL. */
    
    magick_image = NULL;
    magick_image_info = NULL;
    
    capture_window = NULL;
    capture_pixmap = NULL;

    capture_image = NULL;
    capture_gdk_pixmap = NULL;
    capture_gdk_mask = NULL;
    
    /*
     * Determine the file name of the temporary screen capture
     * graphic. ImageMagick saves this graphic onto disk for Imlib
     * to load into a window.
     */
    
    pid = getpid ();
    capture_temp_filename = g_string_new (NULL);
    capture_temp_filename_24 = g_string_new (NULL);
    save_filename = g_string_new (NULL);

    g_string_sprintf
    (
        capture_temp_filename,
        "/tmp/MagiCapture_%d.bmp",
        pid
    );
    g_string_assign
    (
        capture_temp_filename_24,
        capture_temp_filename -> str
    );
    g_string_prepend (capture_temp_filename_24, "BMP24:");

    /* Use Imlib's visual and colormap. */

    visual = gdk_imlib_get_visual ();
    if (visual == NULL)
    {
        error_dialog (ERROR_IMLIB_VISUAL);
        capture_cleanup ();
        exit (-1);
    }
    colormap = gdk_imlib_get_colormap ();
    if (visual == NULL)
    {
        error_dialog (ERROR_IMLIB_COLORMAP);
        capture_cleanup ();
        exit (-1);
    }
    gtk_widget_push_visual (visual);
    gtk_widget_push_colormap (colormap);
}

/*********************************************************************
 *
 * capture_cleanup: Free all memory allocated to screen capture data
 * structures. This function must be called after the GTK main event
 * loop, right before the main function exits.
 *
 * Parameters:
 *   None.
 *
 * Return value:
 *   None.
 */

void capture_cleanup ()
{
    destroy_magick_image ();
    destroy_capture_image ();
    destroy_capture_window ();

    if (capture_temp_filename != NULL)
    {
        unlink (capture_temp_filename -> str);
        g_string_free (capture_temp_filename, TRUE);
        capture_temp_filename = NULL;
    }
    if (capture_temp_filename_24 != NULL)
    {
        g_string_free (capture_temp_filename_24, TRUE);
        capture_temp_filename_24 = NULL;
    }
}

/*********************************************************************
 *
 * capture_set_type: Set the type of entity to capture. The entity
 * may be a single window without decorations, a single window with
 * decorations, or the entire screen.
 *
 * Parameters:
 *   new_cap_type - type of entity to capture
 *
 * Return value:
 *   None.
 */

void capture_set_type (capture_type new_cap_type)
{
    cap_type = new_cap_type;
}

/*********************************************************************
 *
 * capture_set_type: Set the flag that determines whether or not the
 * screen capture window may be visible to the user.
 *
 * Parameters:
 *   new_cap_show - 1 if the window will be shown, 0 otherwise
 *
 * Return value:
 *   None.
 */

void capture_set_show (gint new_cap_show)
{
    cap_show = new_cap_show;
}

/*********************************************************************
 *
 * capture_set_type: Set the delay in seconds between the moment the
 * user clicks on the "Capture!" button and the moment the screen
 * capture process begins.
 *
 * Parameters:
 *   new_cap_delay - delay in seconds before beginning capture process
 *
 * Return value:
 *   None.
 */

void capture_set_delay (gint new_cap_delay)
{
    cap_delay = new_cap_delay;
}

/*********************************************************************
 *
 * capture_do: Proceed with the screen capture!!!!!!!!!!!!!!!!!!!!
 * This function waits a specified number of seconds and calls the
 * ImageMagick XImportImage API function to perform the actual capture.
 * The captured image is saved onto disk into a temporary bitmap file,
 * which will be read later.
 *
 * Parameters:
 *   None.
 *
 * Return value:
 *   _SUCCESS if the operation completed successfully, _FAIL otherwise
 */

gint capture_do ()
{
    unsigned int save_status;

    /* Destroy the existing ImageMagick image structures. */
    
    destroy_magick_image ();

    /* Hide the capture window. */

    if (capture_window != NULL)
    {
        gtk_widget_hide (capture_window);
        gdk_flush ();
    }
    
    /* Initialize the ImageMagick image structures. */

    magick_image_info = CloneImageInfo (NULL);
    GetImageInfo (magick_image_info);
    XGetImportInfo (&magick_import_info);
    magick_import_info.descend = 1;

    /* Set parameters according to the type of display entity to capture. */
    
    switch (cap_type)
    {
        case SINGLE_WINDOW:
            magick_import_info.screen = 0;
            magick_import_info.frame = 0;
            magick_import_info.descend = 1;
            strcpy (magick_image_info -> filename, "");
            break;
        case SINGLE_WINDOW_DECOR:
            magick_import_info.screen = 0;
            magick_import_info.frame = 1;
            magick_import_info.descend = 1;
            strcpy (magick_image_info -> filename, "");
            break;
        case FULL_SCREEN:
            /*
             * TODO: See if the .screen value below should be 1 or 0.
             * We tried 1 before, but it wouldn't capture menus as a
             * user selects them.
             */
            magick_import_info.screen = 0;
            magick_import_info.frame = 1;
            magick_import_info.descend = 1;
            strcpy (magick_image_info -> filename, "root");
            break;
        default:
            /* TODO: Failed assertion here. */
    }

    /* Wait a specified number of seconds. */

    if (cap_delay > 0)
    {
        sleep (cap_delay);
    }
    
    /*
     * Capture the image!!!!!!!!!!!!!!!!!
     *
     * The ImageMagick "import" utility sets the magick parameter of the
     * image structure to "PS", for PostScript, after taking a screen
     * capture. I discovered that this ensures that the image can be
     * saved with the proper number of colors in the palette, particularly
     * if the user wants to save the screen capture as a PNG.
     */

    magick_image = XImportImage (magick_image_info, &magick_import_info);
    strcpy (magick_image -> magick, "PS");  
    if (magick_image == NULL)
    {
        error_dialog (ERROR_MAGICK_IMAGE);
        return _FAIL;
    }

    /* Write the captured image to a temporary bitmap file. */

    strcpy
    (
        magick_image -> filename,
        capture_temp_filename_24 -> str
    );
    strcpy
    (
        magick_image_info -> filename,
        capture_temp_filename_24 -> str
    );
    SetImageInfo (magick_image_info, 1);
    save_status = WriteImage (magick_image_info, magick_image);
    if (!save_status)
    {
        error_dialog (ERROR_SAVE_TEMP);
        return _FAIL;
        /*
         * TODO: Allow the user to be able to save the
         * screen capture without displaying a window.
         */
    }
    return _SUCCESS;
}

/*********************************************************************
 *
 * capture_render: Render the captured image into a GDK window using
 * the Imlib API.
 *
 * Parameters:
 *   None.
 *
 * Return value:
 *   _SUCCESS if the operation completed successfully, _FAIL otherwise
 */

gint capture_render ()
{
    gint w, h;
    gint render_status;
    
    /* Destroy the existing screen capture window. */
    
    destroy_capture_image ();
    destroy_capture_window ();

    /* Load the captured image. */

    capture_image = gdk_imlib_load_image (capture_temp_filename -> str);
    if (capture_image == NULL)
    {
        error_dialog (ERROR_CAPTURE_IMAGE);
        return _FAIL;
    }

    /*
     * Retrieve the image's original width and height
     * out of the Imlib image structure.
     */

    w = capture_image -> rgb_width;
    h = capture_image -> rgb_height;

    /* Render the loaded image into a GDK pixmap. */
    
    render_status = gdk_imlib_render (capture_image, w, h);
    if (render_status == 0)
    {
        error_dialog (ERROR_IMLIB_RENDER);
        return _FAIL;
    }
    capture_gdk_pixmap = gdk_imlib_move_image (capture_image);
    capture_gdk_mask = gdk_imlib_move_mask (capture_image);

    /* Create a window to display the captured image. */

    capture_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    if (capture_window == NULL)
    {
        error_dialog (ERROR_CAPTURE_WINDOW);
        return _FAIL;
    }

    gtk_signal_connect
    (
        GTK_OBJECT (capture_window),
        "delete_event",
        GTK_SIGNAL_FUNC (capture_window_delete_event),
        NULL
    );
    gtk_signal_connect
    (
        GTK_OBJECT (capture_window),
        "destroy",
        GTK_SIGNAL_FUNC (capture_window_destroy),
        NULL
    );
    gtk_container_set_border_width (GTK_CONTAINER (capture_window), 10);
    gtk_window_set_title (GTK_WINDOW (capture_window), TITLE_CAPTURE_WINDOW);

    /*
     * Create a GTK pixmap using the rendered pixmap data and attach it to
     * the capture window.
     */

    capture_pixmap = gtk_pixmap_new (capture_gdk_pixmap, capture_gdk_mask);
    if (capture_pixmap == NULL)
    {
        error_dialog (ERROR_CAPTURE_PIXMAP);
        return _FAIL;
    }
    
    /* TODO: Error check! */
    gtk_container_add (GTK_CONTAINER (capture_window), capture_pixmap);
    gtk_widget_show (capture_pixmap);
    
    gdk_flush ();

    return _SUCCESS;
}

/*********************************************************************
 *
 * capture_show: Show or hide the screen capture window according to
 * whether or not the show radio button is selected.
 *
 * Parameters:
 *   None.
 *
 * Return value:
 *   None.
 */

void capture_show ()
{
    if (capture_window != NULL)
    {
        if (cap_show)
        {
            /* Display the window. */
            gtk_widget_show (capture_window);
        }
        else
        {
            /* Hide the window. */
            gtk_widget_hide (capture_window);
        }
    }
    gdk_flush ();
}

/*********************************************************************
 *
 * capture_save: Save the captured image into a file using
 * a GTK save dialog box.
 *
 * Parameters:
 *   None.
 *
 * Return value:
 *   None.
 */

void capture_save ()
{
    GtkWidget *save_window;

    if (magick_image == NULL)
    {
        error_dialog (ERROR_MAGICK_IMAGE_SAVE);
        return;
    }

    /* Create a GTK save dialog box. */
    
    save_window = gtk_file_selection_new (TITLE_SAVE_WINDOW);

    /* Connect the OK and Cancel buttons to the appropriate operations. */

    gtk_signal_connect
    (
        GTK_OBJECT (GTK_FILE_SELECTION (save_window) -> ok_button),
        "clicked",
        (GtkSignalFunc) capture_save_ok,
        save_window
    );
    gtk_signal_connect_object
    (
        GTK_OBJECT (GTK_FILE_SELECTION (save_window) -> ok_button),
        "clicked",
        (GtkSignalFunc) gtk_widget_destroy,
        GTK_OBJECT (save_window)
    );
    gtk_signal_connect_object
    (
        GTK_OBJECT (GTK_FILE_SELECTION (save_window) -> cancel_button),
        "clicked",
        (GtkSignalFunc) gtk_widget_destroy,
        GTK_OBJECT (save_window)
    );

    /* Set the default filename to what was used previously. */
    
    if (save_filename != NULL)
    {
        gtk_file_selection_set_filename
        (
            GTK_FILE_SELECTION (save_window),
            save_filename -> str
        );
    }
    
    /* Show the save dialog box. */

    gtk_widget_show (save_window);
}

/*********************************************************************
 *
 * capture_save_ok: Once the OK button is clicked on the save dialog
 * box, save the captured image into a file using ImageMagick's save
 * capability.
 *
 * Parameters:
 *   w - the GTK widget involved
 *   fs - the save dialog box
 *
 * Return value:
 *   None.
 */

void capture_save_ok (GtkWidget *w, GtkFileSelection *fs)
{
    unsigned int save_status;
    gchar *save_filename_text;

    save_filename_text =
    gtk_file_selection_get_filename (GTK_FILE_SELECTION (fs));
    g_string_assign (save_filename, save_filename_text);
    g_print ("Saving file %s\n", save_filename -> str);

    /*
     * TODO: On certain PseudoColor displays, ImageMagick
     * complains that the number of colors in the image is invalid
     * while trying to save to a PNG file. We need to fix this.
     */
    
    strcpy (magick_image -> filename, save_filename -> str);
    strcpy (magick_image_info -> filename, save_filename -> str);
    SetImageInfo (magick_image_info, 1);
    save_status = WriteImage (magick_image_info, magick_image);
    if (!save_status)
    {
        error_dialog (ERROR_SAVE);
        return;
    }
}
