/*
 * Copyright (C) 1999 Arthur Jerijian
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * In addition, permission is hereby granted, free of charge, to any person
 * obtaining a copy of the software, to deal in the Software according to the
 * terms outlined in the GNU General Public License (the "GPL") in lieu of the
 * licensing terms stated in the above permission notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) OF THE SOFTWARE BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * panel.c: Routines for managing the widgets on the main panel
 */

#include "common.h"
#include "capture.h"
#include "stringtable.h"
#include "panel.h"

/*********************************************************************
 *
 * add_button: Add a button to the main panel.
 *
 * Parameters:
 *   box - the box where the button will be added
 *   button_func - the button's callback function
 *   button_text - the text label to place within the button
 *
 * Return value:
 *   None.
 */

void add_button
(
    GtkWidget *box,
    GtkSignalFunc button_func,
    const gchar *button_text
)
{
    GtkWidget *button;
    
    button = gtk_button_new_with_label (button_text);
    gtk_box_pack_start (GTK_BOX (box), button, FALSE, FALSE, 0);
    gtk_signal_connect
    (
        GTK_OBJECT (button),
        "clicked",
        GTK_SIGNAL_FUNC (button_func),
        NULL
    );
    gtk_widget_show (button);
}

/*********************************************************************
 *
 * add_radio: Add a radio button to the main panel.
 *
 * Parameters:
 *   box - the box where the radio button will be added
 *   group - the group affected when the user selects the radio button
 *   radio_func - the radio button's callback function
 *   active - whether to set the radio button active by default
 *   radio_text - the text label to place within the radio button
 *
 * Return value:
 *   The group containing the newly created radio button
 */

GSList *add_radio
(
    GtkWidget *box,
    GSList *group,
    GtkSignalFunc radio_func,
    gint active,
    const gchar *radio_text
)
{
    GtkWidget *radio;
    
    radio = gtk_radio_button_new_with_label (group, radio_text);
    gtk_box_pack_start (GTK_BOX (box), radio, TRUE, TRUE, 0);
    gtk_widget_show (radio);
    
    gtk_toggle_button_set_active
    (
        GTK_TOGGLE_BUTTON (radio),
        active
    );
    
    if (radio_func != NULL)
    {
        gtk_signal_connect
        (
            GTK_OBJECT (radio),
            "toggled",
            GTK_SIGNAL_FUNC (radio_func),
            NULL
        );
    }
    return (gtk_radio_button_group (GTK_RADIO_BUTTON (radio)));
}

/*********************************************************************
 *
 * add_separator: Add a separator to the main panel.
 *
 * Parameters:
 *   box - the box where the separator will be added
 *
 * Return value:
 *   None.
 */

void add_separator (GtkWidget *box)
{
    GtkWidget *separator;
    
    separator = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (box), separator, FALSE, TRUE, 5);
    gtk_widget_show (separator);
}

/*********************************************************************
 *
 * add_label: Add a text label to the main panel.
 *
 * Parameters:
 *   box - the box where the text label will be added
 *   label_text - the string to display within the text label
 *
 * Return value:
 *   None.
 */

void add_label (GtkWidget *box, const gchar *label_text)
{
    GtkWidget *label;
    
    label = gtk_label_new (label_text);
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0);
    gtk_box_pack_start (GTK_BOX (box), label, TRUE, TRUE, 5);
    gtk_widget_show (label);
}

/*********************************************************************
 *
 * add_delay: Add the delay parameter to the main panel. The delay
 * parameter consists of a text box where the user enters the number
 * of seconds to wait before retrieving a screen capture, followed by
 * the internationalized string for the word "seconds".
 *
 * Parameters:
 *   box - the box where the separator will be added
 *
 * Return value:
 *   None.
 */

void add_delay (GtkWidget *box)
{
    GtkWidget *box_delay;
    GtkWidget *entry_delay;

    /* Both new widgets will be placed within a new hbox. */
    
    box_delay = gtk_hbox_new (FALSE, 0);

    /* Create the delay parameter text box. */
    
    entry_delay = gtk_entry_new_with_max_length (DELAY_TEXTBOX_SIZE);
    gtk_entry_set_text (GTK_ENTRY (entry_delay), DEFAULT_DELAY_STR);
    gtk_widget_set_usize (entry_delay, 50, 0);
    gtk_signal_connect
    (
        GTK_OBJECT (entry_delay),
        "changed",
        GTK_SIGNAL_FUNC (delay_func),
        NULL
    );
    gtk_box_pack_start (GTK_BOX (box_delay), entry_delay, FALSE, FALSE, 0);
    gtk_widget_show (entry_delay);

    /* Create the following text label. */
    
    add_label (box_delay, LABEL_SECONDS);
    
    gtk_box_pack_start (GTK_BOX (box), box_delay, FALSE, FALSE, 5);
    gtk_widget_show (box_delay);
}

/*********************************************************************
 *
 * create_main_panel: Create the main panel widget which includes
 * all buttons, labels, and text boxes belonging to the main panel.
 *
 * Parameters:
 *   None.
 *
 * Return value:
 *   None.
 */

GtkWidget *create_main_panel ()
{
    GtkWidget *box;
    GSList *group;
    
    box = gtk_vbox_new (FALSE, 0);

    /*
     * Add radio buttons for capture mode: single window,
     * single window with decorations, and full screen.
     */
    
    add_label (box, LABEL_CAPTURE);
    group = add_radio
    (
        box,
        NULL,
        GTK_SIGNAL_FUNC (radio_single_func),
        TRUE,
        LABEL_SINGLE_WINDOW
    );
    group = add_radio
    (
        box,
        group,
        GTK_SIGNAL_FUNC (radio_single_decor_func),
        FALSE,
        LABEL_SINGLE_WINDOW_DECOR
    );
    group = add_radio
    (
        box,
        group,
        GTK_SIGNAL_FUNC (radio_full_func),
        FALSE,
        LABEL_FULL_SCREEN
    );

    /* Add delay parameter widgets. */
    
    add_label (box, LABEL_DELAY);
    add_delay (box);

    /*
     * Add radio buttons for showing and hiding the capture
     * preview window.
     */
    
    add_label (box, LABEL_VIEW);
    group = add_radio
    (
        box,
        NULL,
        GTK_SIGNAL_FUNC (radio_show_func),
        TRUE,
        LABEL_SHOW_CAPTURE
    );
    group = add_radio
    (
        box,
        group,
        GTK_SIGNAL_FUNC (radio_hide_func),
        FALSE,
        LABEL_HIDE_CAPTURE
    );
    
    add_separator (box);

    /* Add the capture button. */
    
    add_button
    (
        box,
        GTK_SIGNAL_FUNC (button_capture_func),
        LABEL_CAPTURE_DO
    );
    
    add_separator (box);

    /*
     * Add buttons for configuring options, saving the capture to
     * a file, and quitting the program.
     */
    
    add_button
    (
        box,
        GTK_SIGNAL_FUNC (button_options_func),
        LABEL_OPTIONS
    );
    add_button
    (
        box,
        GTK_SIGNAL_FUNC (button_saveas_func),
        LABEL_SAVEAS
    );
    add_button
    (
        box,
        GTK_SIGNAL_FUNC (button_quit_func),
        LABEL_QUIT
    );
    
    return (box);
}

/*********************************************************************
 *
 * radio_single_func: GTK callback function called when the user
 * selects the "Single Window" radio button.
 *
 * Parameters:
 *   widget - GTK widget connected to the callback function
 *   data - callback function data
 *
 * Return value:
 *   None.
 */

void radio_single_func (GtkWidget *widget, gpointer data)
{
    if (GTK_TOGGLE_BUTTON (widget) -> active)
    {
        debug_print (LABEL_SINGLE_WINDOW);
        capture_set_type (SINGLE_WINDOW);
    }
}

/*********************************************************************
 *
 * radio_single_decor_func: GTK callback function called when the user
 * selects the "Single Window with Decorations" radio button.
 *
 * Parameters:
 *   widget - GTK widget connected to the callback function
 *   data - callback function data
 *
 * Return value:
 *   None.
 */

void radio_single_decor_func (GtkWidget *widget, gpointer data)
{
    if (GTK_TOGGLE_BUTTON (widget) -> active)
    {
        debug_print (LABEL_SINGLE_WINDOW_DECOR);
        capture_set_type (SINGLE_WINDOW_DECOR);
    }
}

/*********************************************************************
 *
 * radio_full_func: GTK callback function called when the user
 * selects the "Full Screen" radio button.
 *
 * Parameters:
 *   widget - GTK widget connected to the callback function
 *   data - callback function data
 *
 * Return value:
 *   None.
 */

void radio_full_func (GtkWidget *widget, gpointer data)
{
    if (GTK_TOGGLE_BUTTON (widget) -> active)
    {
        debug_print (LABEL_FULL_SCREEN);
        capture_set_type (FULL_SCREEN);
    }
}

/*********************************************************************
 *
 * radio_show_func: GTK callback function called when the user
 * selects the "Show Capture" radio button.
 *
 * Parameters:
 *   widget - GTK widget connected to the callback function
 *   data - callback function data
 *
 * Return value:
 *   None.
 */

void radio_show_func (GtkWidget *widget, gpointer data)
{
    if (GTK_TOGGLE_BUTTON (widget) -> active)
    {
        debug_print (LABEL_SHOW_CAPTURE);
        capture_set_show (1);
        capture_show ();
    }
}

/*********************************************************************
 *
 * radio_hide_func: GTK callback function called when the user
 * selects the "Hide Capture" radio button.
 *
 * Parameters:
 *   widget - GTK widget connected to the callback function
 *   data - callback function data
 *
 * Return value:
 *   None.
 */

void radio_hide_func (GtkWidget *widget, gpointer data)
{
    if (GTK_TOGGLE_BUTTON (widget) -> active)
    {
        debug_print (LABEL_HIDE_CAPTURE);
        capture_set_show (0);
        capture_show ();
    }
}

/*********************************************************************
 *
 * button_capture_func: GTK callback function called when the user
 * clicks on the "Capture!" button.
 *
 * Parameters:
 *   widget - GTK widget connected to the callback function
 *   data - callback function data
 *
 * Return value:
 *   None.
 */

void button_capture_func (GtkWidget *widget, gpointer data)
{
    gint cap_status;
    
    debug_print (LABEL_CAPTURE_DO);
        
    /* TODO: Possibly hide the main control window here... */
    
    cap_status = capture_do ();
    if (cap_status == _FAIL) { return; }
    cap_status = capture_render ();
    if (cap_status == _FAIL) { return; }
    
    capture_show ();
    
    /* TODO: Possibly show the main window again ... */
}

/*********************************************************************
 *
 * button_options_func: GTK callback function called when the user
 * clicks on the "Options..." button.
 *
 * Parameters:
 *   widget - GTK widget connected to the callback function
 *   data - callback function data
 *
 * Return value:
 *   None.
 */

void button_options_func (GtkWidget *widget, gpointer data)
{
    debug_print (LABEL_OPTIONS);
    error_dialog (ERROR_NOT_IMPLEMENTED);
    
    /* TODO: We need code here....... */
}

/*********************************************************************
 *
 * button_saveas_func: GTK callback function called when the user
 * clicks on the "Save As..." button.
 *
 * Parameters:
 *   widget - GTK widget connected to the callback function
 *   data - callback function data
 *
 * Return value:
 *   None.
 */

void button_saveas_func (GtkWidget *widget, gpointer data)
{
    debug_print (LABEL_SAVEAS);
    capture_save ();
}

/*********************************************************************
 *
 * button_quit_func: GTK callback function called when the user
 * clicks on the "Quit" button.
 *
 * Parameters:
 *   widget - GTK widget connected to the callback function
 *   data - callback function data
 *
 * Return value:
 *   None.
 */

void button_quit_func (GtkWidget *widget, gpointer data)
{
    debug_print (LABEL_QUIT);
    gtk_main_quit ();
}

/*********************************************************************
 *
 * delay_func: GTK callback function called when the user
 * modifies the contents of the delay parameter text box.
 *
 * Parameters:
 *   widget - GTK widget connected to the callback function
 *   data - callback function data
 *
 * Return value:
 *   None.
 */

void delay_func (GtkWidget *widget, gpointer data)
{
    static unsigned int delay = DEFAULT_DELAY;
    GString *prev_entered;
    int i;
    gchar *entered;
    GString *delay_debug_buffer;
    
    entered = gtk_entry_get_text (GTK_ENTRY (widget));

    /*
     * Ensure that the delay parameter text box contains
     * only numerals. Currently, the procedure works well
     * with the ASCII character set, but not with the EBCDIC
     * character set (which may not even be used on UNIX).
     */
    
    for (i = 0; entered [i] != '\0'; i++)
    {
        if ((entered [i] < '0') || (entered [i] > '9'))
        {
            prev_entered = g_string_new (NULL);
            g_string_sprintf (prev_entered, "%d", delay);
            gtk_entry_set_text (GTK_ENTRY (widget), prev_entered -> str);
            gtk_entry_set_position
            (
                GTK_ENTRY (widget),
                prev_entered -> len - 1
            );
            g_string_free (prev_entered, TRUE);
            return;
        }
    }
    delay = atoi (entered);
    capture_set_delay (delay);

#ifdef MAGICAPTURE_DEBUG    
    delay_debug_buffer = g_string_new (NULL);
    g_string_sprintf
    (
        delay_debug_buffer,
        "%s%d%s",
        LABEL_DELAY,
        delay,
        LABEL_SECONDS
    );
    debug_print (delay_debug_buffer -> str);
    g_string_free (delay_debug_buffer, TRUE);
#endif /* MAGICAPTURE_DEBUG */
}
